@@ -17,11 +17,14 @@
 System-level utilities and helper functions.
 """
 
+import copy
 import math
 import re
 import unicodedata
 
+import pyparsing as pp
 import six
+from six.moves import urllib
 
 from oslo_utils._i18n import _
 from oslo_utils import encodeutils
@@ -41,6 +44,7 @@ UNIT_PREFIX_EXPONENT = {
 UNIT_SYSTEM_INFO = {
     'IEC': (1024, re.compile(r'(^[-+]?\d*\.?\d+)([KMGT]i?)?(b|bit|B)$')),
     'SI': (1000, re.compile(r'(^[-+]?\d*\.?\d+)([kMGT])?(b|bit|B)$')),
+    'mixed': (None, re.compile(r'(^[-+]?\d*\.?\d+)([kKMGT]i?)?(b|bit|B)$')),
 }
 
 TRUE_STRINGS = ('1', 't', 'true', 'on', 'y', 'yes')
@@ -53,7 +57,8 @@ SLUGIFY_HYPHENATE_RE = re.compile(r"[-\s]+")
 # NOTE(flaper87): The following globals are used by `mask_password`
 _SANITIZE_KEYS = ['adminPass', 'admin_pass', 'password', 'admin_password',
                   'auth_token', 'new_pass', 'auth_password', 'secret_uuid',
-                  'secret', 'sys_pswd', 'token', 'configdrive']
+                  'secret', 'sys_pswd', 'token', 'configdrive',
+                  'CHAPPASSWORD', 'encrypted_key', 'private_key']
 
 # NOTE(ldbragst): Let's build a list of regex objects using the list of
 # _SANITIZE_KEYS we already have. This way, we only have to add the new key
@@ -107,11 +112,14 @@ def int_from_bool_as_string(subject):
 
 
 def bool_from_string(subject, strict=False, default=False):
-    """Interpret a string as a boolean.
+    """Interpret a subject as a boolean.
 
-    A case-insensitive match is performed such that strings matching 't',
-    'true', 'on', 'y', 'yes', or '1' are considered True and, when
-    `strict=False`, anything else returns the value specified by 'default'.
+    A subject can be a boolean, a string or an integer. Boolean type value
+    will be returned directly, otherwise the subject will be converted to
+    a string. A case-insensitive match is performed such that strings
+    matching 't','true', 'on', 'y', 'yes', or '1' are considered True and,
+    when `strict=False`, anything else returns the value specified by
+    'default'.
 
     Useful for JSON-decoded stuff and config file parsing.
 
@@ -141,10 +149,23 @@ def bool_from_string(subject, strict=False, default=False):
         return default
 
 
+def is_valid_boolstr(value):
+    """Check if the provided string is a valid bool string or not.
+
+    :param value: value to verify
+    :type value: string
+    :returns: true if value is boolean string, false otherwise
+
+    .. versionadded:: 3.17
+    """
+    boolstrs = TRUE_STRINGS + FALSE_STRINGS
+    return str(value).lower() in boolstrs
+
+
 def string_to_bytes(text, unit_system='IEC', return_int=False):
     """Converts a string into an float representation of bytes.
 
-    The units supported for IEC ::
+    The units supported for IEC / mixed::
 
         Kb(it), Kib(it), Mb(it), Mib(it), Gb(it), Gib(it), Tb(it), Tib(it)
         KB, KiB, MB, MiB, GB, GiB, TB, TiB
@@ -154,7 +175,17 @@ def string_to_bytes(text, unit_system='IEC', return_int=False):
         kb(it), Mb(it), Gb(it), Tb(it)
         kB, MB, GB, TB
 
-    Note that the SI unit system does not support capital letter 'K'
+    SI units are interpreted as power-of-ten (e.g. 1kb = 1000b).  Note
+    that the SI unit system does not support capital letter 'K'
+
+    IEC units are interpreted as power-of-two (e.g. 1MiB = 1MB =
+    1024b)
+
+    Mixed units interpret the "i" to mean IEC, and no "i" to mean SI
+    (e.g. 1kb = 1000b, 1kib == 1024b).  Additionaly, mixed units
+    interpret 'K' as power-of-ten.  This mode is not particuarly
+    useful for new code, but can help with compatability for parsers
+    such as GNU parted.
 
     :param text: String input for bytes size conversion.
     :param unit_system: Unit system for byte size conversion.
@@ -175,9 +206,22 @@ def string_to_bytes(text, unit_system='IEC', return_int=False):
         unit_prefix = match.group(2)
         if match.group(3) in ['b', 'bit']:
             magnitude /= 8
+
+        # In the mixed matcher, IEC units (with a trailing 'i') are
+        # interpreted as power-of-two, others as power-of-ten
+        if unit_system == 'mixed':
+            if unit_prefix and not unit_prefix.endswith('i'):
+                # For maximum compatability in mixed mode, we understand
+                # "K" (which is not strict SI) as "k"
+                if unit_prefix.startswith == 'K':
+                    unit_prefix = 'k'
+                base = 1000
+            else:
+                base = 1024
     else:
         msg = _('Invalid string format: %s') % text
         raise ValueError(msg)
+
     if not unit_prefix:
         res = magnitude
     else:
@@ -265,6 +309,9 @@ def mask_password(message, secret="***"):  # nosec
 
     .. versionchanged:: 3.4
        Replace also ``'configdrive'`` key.
+
+    .. versionchanged:: 3.8
+       Replace also ``'CHAPPASSWORD'`` key.
     """
 
     try:
@@ -300,16 +347,16 @@ def mask_dict_password(dictionary, secret="***"):  # nosec
     A dictionary (which may contain nested dictionaries) contains
     information (such as passwords) which should not be revealed, and
     this function helps detect and replace those with the 'secret'
-    provided (or '***' if none is provided).
+    provided (or `***` if none is provided).
 
     Substitution is performed in one of three situations:
 
     If the key is something that is considered to be indicative of a
     secret, then the corresponding value is replaced with the secret
-    provided (or '***' if none is provided).
+    provided (or `***` if none is provided).
 
     If a value in the dictionary is a string, then it is masked
-    using the mask_password() function.
+    using the ``mask_password()`` function.
 
     Finally, if a value is a dictionary, this function will
     recursively mask that dictionary as well.
@@ -348,18 +395,29 @@ def mask_dict_password(dictionary, secret="***"):  # nosec
         raise TypeError("Expected a dictionary, got %s instead."
                         % type(dictionary))
 
-    out = {}
+    out = copy.deepcopy(dictionary)
 
     for k, v in dictionary.items():
         if isinstance(v, dict):
-            v = mask_dict_password(v, secret=secret)
-        elif k in _SANITIZE_KEYS:
-            v = secret
-        elif isinstance(v, six.string_types):
-            v = mask_password(v, secret=secret)
-
-        out[k] = v
-
+            out[k] = mask_dict_password(v, secret=secret)
+            continue
+        # NOTE(jlvillal): Check to see if anything in the dictionary 'key'
+        # contains any key specified in _SANITIZE_KEYS.
+        k_matched = False
+        if isinstance(k, six.string_types):
+            for sani_key in _SANITIZE_KEYS:
+                if sani_key in k:
+                    out[k] = secret
+                    k_matched = True
+                    break
+        if not k_matched:
+            # We did not find a match for the key name in the
+            # _SANITIZE_KEYS, so we fall through to here
+            if isinstance(v, six.string_types):
+                out[k] = mask_password(v, secret=secret)
+            else:
+                # Just leave it alone.
+                out[k] = v
     return out
 
 
@@ -408,3 +466,101 @@ def check_string_length(value, name=None, min_length=0, max_length=None):
                 "%(max_length)s.") % {'name': name, 'length': length,
                                       'max_length': max_length}
         raise ValueError(msg)
+
+
+def validate_integer(value, name, min_value=None, max_value=None):
+    """Make sure that value is a valid integer, potentially within range.
+
+    :param value: value of the integer
+    :param name: name of the integer
+    :param min_value: min_value of the integer
+    :param max_value: max_value of the integer
+    :returns: integer
+    :raises: ValueError if value is an invalid integer
+
+    .. versionadded:: 3.33
+    """
+    try:
+        value = int(str(value))
+    except (ValueError, UnicodeEncodeError):
+        msg = _('%(value_name)s must be an integer'
+                ) % {'value_name': name}
+        raise ValueError(msg)
+
+    if min_value is not None and value < min_value:
+        msg = _('%(value_name)s must be >= %(min_value)d'
+                ) % {'value_name': name, 'min_value': min_value}
+        raise ValueError(msg)
+
+    if max_value is not None and value > max_value:
+        msg = _('%(value_name)s must be <= %(max_value)d'
+                ) % {'value_name': name, 'max_value': max_value}
+        raise ValueError(msg)
+
+    return value
+
+
+def split_path(path, minsegs=1, maxsegs=None, rest_with_last=False):
+    """Validate and split the given HTTP request path.
+
+    **Examples**::
+
+        ['a'] = _split_path('/a')
+        ['a', None] = _split_path('/a', 1, 2)
+        ['a', 'c'] = _split_path('/a/c', 1, 2)
+        ['a', 'c', 'o/r'] = _split_path('/a/c/o/r', 1, 3, True)
+
+    :param path: HTTP Request path to be split
+    :param minsegs: Minimum number of segments to be extracted
+    :param maxsegs: Maximum number of segments to be extracted
+    :param rest_with_last: If True, trailing data will be returned as part
+                           of last segment.  If False, and there is
+                           trailing data, raises ValueError.
+    :returns: list of segments with a length of maxsegs (non-existent
+              segments will return as None)
+    :raises: ValueError if given an invalid path
+
+    .. versionadded:: 3.11
+    """
+    if not maxsegs:
+        maxsegs = minsegs
+    if minsegs > maxsegs:
+        raise ValueError(_('minsegs > maxsegs: %(min)d > %(max)d)') %
+                         {'min': minsegs, 'max': maxsegs})
+    if rest_with_last:
+        segs = path.split('/', maxsegs)
+        minsegs += 1
+        maxsegs += 1
+        count = len(segs)
+        if (segs[0] or count < minsegs or count > maxsegs or
+                '' in segs[1:minsegs]):
+            raise ValueError(_('Invalid path: %s') % urllib.parse.quote(path))
+    else:
+        minsegs += 1
+        maxsegs += 1
+        segs = path.split('/', maxsegs)
+        count = len(segs)
+        if (segs[0] or count < minsegs or count > maxsegs + 1 or
+                '' in segs[1:minsegs] or
+                (count == maxsegs + 1 and segs[maxsegs])):
+            raise ValueError(_('Invalid path: %s') % urllib.parse.quote(path))
+    segs = segs[1:maxsegs]
+    segs.extend([None] * (maxsegs - 1 - len(segs)))
+    return segs
+
+
+def split_by_commas(value):
+    """Split values by commas and quotes according to api-wg
+
+    :param value: value to be split
+
+    .. versionadded:: 3.17
+    """
+    word = (pp.QuotedString(quoteChar='"', escChar='\\')
+            | pp.Word(pp.printables, excludeChars='",'))
+    grammar = pp.stringStart + pp.delimitedList(word) + pp.stringEnd
+
+    try:
+        return list(grammar.parseString(value))
+    except pp.ParseException:
+        raise ValueError("Invalid value: %s" % value)
